""" RHEL download """
import os

from virt.qemutest.distro.core import FetchDistroCore, FetchNetinst

def rhel_toplevel(release, nightly):
    major = release.split('.')[0]
    group = 'rel-eng'
    subdir = 'BaseOS'
    if nightly:
        group = 'nightly'
    if int(major) < 8:
        subdir = 'Server'
    return f'rhel-{major}/{group}/RHEL-{major}/latest-RHEL-{release}/compose/{subdir}'

class FetchRHELCloud(FetchDistroCore):
    """ class for rhel cloud image download """

    def __init__(self, mirror, arch = None, nightly = False):
        super().__init__(arch)
        self.mirror = mirror
        self.nightly = nightly

    def baseurl(self, release):
        toplevel = rhel_toplevel(release, self.nightly)
        return f'{self.mirror}/{toplevel}/{self.arch}/images/'

    def image_id(self, release):
        return f'rhel-{release}-{self.arch}-cloud'

    def find_image(self, release, pattern):
        checksum = f'{self.baseurl(release)}SHA256SUM'
        return self.checksum_find_latest(checksum, pattern)

    def fetch_image(self, release, pattern = 'guest-image', imgdir = 'distro'):
        print(f'# ----- rhel {release} ({pattern}) -----')
        (latest, checksum) = self.find_image(release, pattern)
        url = f'{self.baseurl(release)}{latest}'
        out = f'{imgdir}/{latest}'
        os.makedirs(imgdir, exist_ok = True)
        self.fetch_image_checksum(url, out, checksum)
        self.index_update(imgdir, self.image_id(release), latest, checksum)

class FetchRHELInstall(FetchDistroCore):
    """ class for rhel full install iso download """

    def __init__(self, mirror, arch = None, nightly = False):
        super().__init__(arch)
        self.mirror = mirror
        self.nightly = nightly

    def baseurl(self, release):
        toplevel = rhel_toplevel(release, self.nightly)
        return f'{self.mirror}/{toplevel}/{self.arch}/iso/'

    def image_id(self, release):
        return f'rhel-{release}-{self.arch}-install'

    def find_image(self, release, pattern):
        checksum = f'{self.baseurl(release)}SHA256SUM'
        return self.checksum_find_latest(checksum, pattern)

    def fetch_image(self, release, pattern = 'dvd', imgdir = 'distro'):
        print(f'# ----- rhel {release} ({pattern}) -----')
        (latest, checksum) = self.find_image(release, pattern)
        url = f'{self.baseurl(release)}{latest}'
        out = f'{imgdir}/{latest}'
        os.makedirs(imgdir, exist_ok = True)
        self.fetch_image_checksum(url, out, checksum)
        self.index_update(imgdir, self.image_id(release), latest, checksum,
                          url = url)

class FetchRHELNetinst(FetchNetinst):
    """ class for rhel netinst iso download """

    nightly = False

    def __init__(self, mirror, arch = None, nightly = False):
        super().__init__(mirror, arch)
        self.nightly = nightly

    def baseurl(self, release):
        toplevel = rhel_toplevel(release, self.nightly)
        return f'{self.mirror}/{toplevel}/{self.arch}/os/'



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2026 Gerd Hoffmann
#
# pylint: disable=consider-using-with
""" boot cloud images on qemu """
import os
import sys
import shutil
import argparse
import tempfile
import subprocess

from virt.qemutest.qemucfg import QemuCfg
from virt.qemutest.autoconf import find_qemu_binary, find_firmware, qemu_default_machine
from virt.qemutest.cloudinit import CloudInitISO
from virt.qemutest.distro.core import default_imgdir, DistroList

def xterm_title(msg):
    term = os.environ.get('TERM', 'dumb')
    if term.startswith('xterm'):
        print(f'\x1b]2;{msg}\x07', flush = True, end = '')

def find_cloud_image(imgdir, name, release):
    l = DistroList(imgdir)
    image_id = l.mkid(f'{name}-{release}', 'cloud')
    image = l.find_filename(image_id)
    if image is None:
        print(f"ERROR: no {name}-{release} cloud image found")
        print("HINT: try virt-cloud-fetch")
        return None
    if os.path.exists(image):
        return image
    print(f"ERROR: file {image} does not exist")
    return None

# pylint: disable=too-many-statements
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-branches
def main():
    parser = argparse.ArgumentParser(description = "boot cloud image")
    parser.add_argument('--dry-run', '--dryrun', dest = 'dryrun',
                        action = 'store_true', default = False,
                        help = 'print generated qemu command line and quit')

    defdir = default_imgdir()
    pgroup = parser.add_argument_group('select cloud image')
    pgroup.add_argument('-i', '--image', dest = 'image', type = str,
                        help = 'boot cloud image FILE', metavar = 'FILE')
    pgroup.add_argument('--fedora', dest = 'fedora', type = str,
                        help = 'boot fedora VER', metavar = 'VER')
    pgroup.add_argument('--fedora-uki', dest = 'fedora_uki', type = str,
                        help = 'boot fedora VER (uki variant)', metavar = 'VER')
    pgroup.add_argument('--centos', dest = 'centos', type = str,
                        help = 'boot centos stream VER', metavar = 'VER')
    pgroup.add_argument('--rhel', dest = 'rhel', type = str,
                        help = 'boot rhel VER', metavar = 'VER')
    pgroup.add_argument('--imgdir', dest = 'imgdir',
                        type = str, default = defdir,
                        help = f'lookup fedora/centos/rhel images in DIR (default: {defdir})',
                        metavar = 'DIR')

    pgroup = parser.add_argument_group('setup cloud image')
    pgroup.add_argument('-u', '--username', '--user', dest = 'userid', default = 'cloud',
                        help = 'login username (default: cloud)', metavar = 'USER')
    pgroup.add_argument('--password', '--passwd', dest = 'passwd', default = 'cloud',
                        help = 'login password (default: cloud)', metavar = 'PASSWD')
    pgroup.add_argument('--package', '--rpm', dest = 'package',
                        action = 'append', type = str, default = [],
                        help = 'install package PKG', metavar = 'PKG')
    pgroup.add_argument('--copy-file', dest = 'copyfiles',
                        action = 'append', type = str, default = [],
                        help = 'copy FILE into the guest', metavar = 'FILE')
    pgroup.add_argument('--runcmd', dest = 'runcmd',
                        action = 'append', type = str, default = [],
                        help = 'run CMD in guest', metavar = 'CMD')

    pgroup = parser.add_argument_group('firmware setup')
    pgroup.add_argument('--smm', dest = 'smm',
                        action = 'store_true', default = False,
                        help = 'use smm firmware builds (feature "require-smm")')
    pgroup.add_argument('--qemu-vars', '--qemuvars', dest = 'qemuvars',
                        action = 'store_true', default = False,
                        help = 'use qemu uefi vars (feature "host-uefi-vars")')
    pgroup.add_argument('--secure-boot', '--sb', dest = 'sb',
                        action = 'store_true', default = False,
                        help = 'enable secure boot (feature "enrolled-keys")')
    pgroup.add_argument('--cc-mode', dest = 'cc_mode',
                        choices=['none', 'sev', 'snp', 'tdx'], default = 'none',
                        help = 'set confidential computing mode (default: none)')
    pgroup.add_argument('--sev', dest = 'cc_mode',
                        action = 'store_const', const = 'sev',
                        help = 'enable sev cc mode (feature "amd-sev")')
    pgroup.add_argument('--snp', dest = 'cc_mode',
                        action = 'store_const', const = 'snp',
                        help = 'enable snp cc mode (feature "amd-sev-snp")')
    pgroup.add_argument('--tdx', dest = 'cc_mode',
                        action = 'store_const', const = 'tdx',
                        help = 'enable tdx cc mode (feature "intel-tdx")')
    pgroup.add_argument('--log', dest = 'log',
                        help = 'write firmware log to FILE', metavar = 'FILE')

    pgroup = parser.add_argument_group('hardware setup')
    pgroup.add_argument('--memory', '--mem', dest = 'memory', default = '4G',
                        help = 'assign guest memory (default: 4G)', metavar = 'MEM')
    pgroup.add_argument('--smp', dest = 'smp', default = '1',
                        help = 'assign guest vcpus (default: 1)', metavar = 'CPUS')
    pgroup.add_argument('--scsi', dest = 'use_scsi',
                        action = argparse.BooleanOptionalAction, default = True,
                        help = 'use virtio scsi (default: yes)')
    pgroup.add_argument('--sata', dest = 'use_sata',
                        action = argparse.BooleanOptionalAction, default = False,
                        help = 'use sata (q35 only, default: no)')
    pgroup.add_argument('--nvme', dest = 'use_nvme',
                        action = argparse.BooleanOptionalAction, default = False,
                        help = 'use NVMe emulation (default: no)')

    options = parser.parse_args()

    # basics
    qemu = find_qemu_binary()
    if qemu is None:
        print("ERROR: no qemu binary found")
        return 1
    machine = qemu_default_machine()
    cmdline = [ qemu, ]
    cmdline += [ '-machine', machine ]
    cmdline += [ '-m', options.memory ]
    cmdline += [ '-smp', options.smp ]
    cmdline += [ '-cpu', 'host' ]
    cmdline += [ '-enable-kvm', ]
    cmdline += [ '-nographic', ]

    # qemucfg
    use_sata = False
    if machine.startswith('q35') and options.use_sata:
        use_sata = True
    qemucfg = QemuCfg(scsi = options.use_scsi,
                      nvme = options.use_nvme,
                      sata = use_sata)

    # firmware
    fw = find_firmware(sb = options.sb,
                       smm = options.smm,
                       qemuvars = options.qemuvars,
                       sev = options.cc_mode == 'sev',
                       snp = options.cc_mode == 'snp',
                       tdx = options.cc_mode == 'tdx')

    if fw is None:
        print("ERROR: no firmware found")
        return 1
    template = fw.vars_template()
    if template:
        varsfile = tempfile.NamedTemporaryFile()
        shutil.copyfile(template, varsfile.name)
        cmdline += fw.cmdline(varsfile = varsfile.name)
    else:
        varsfile = None
        cmdline += fw.cmdline()
    qemucfg.add_confidential(options.cc_mode)

    # hardware
    qemucfg.add_virtio_rng()
    qemucfg.add_network()
    if options.log and os.uname().machine == 'x86_64':
        qemucfg.add_x64_firmware_log(options.log)

    # cloud image
    if options.image:
        hostname = 'cloud-image'
        title = options.image
        image = options.image
    elif options.rhel:
        hostname = f'rhel-{options.rhel}'
        title = f'RHEL {options.rhel}'
        image = find_cloud_image(options.imgdir, 'rhel', options.rhel)
        if image is None:
            return 1
    elif options.centos:
        hostname = f'centos-{options.centos}'
        title = f'CentOS Stream {options.centos}'
        image = find_cloud_image(options.imgdir, 'centos-stream', options.centos)
        if image is None:
            return 1
    elif options.fedora:
        hostname = f'fedora-{options.fedora}'
        title = f'Fedora {options.fedora}'
        image = find_cloud_image(options.imgdir, 'fedora', options.fedora)
        if image is None:
            return 1
    elif options.fedora_uki:
        hostname = f'fedora-uki-{options.fedora_uki}'
        title = f'Fedora {options.fedora} (UKI)'
        image = find_cloud_image(options.imgdir, 'fedora-uki', options.fedora_uki)
        if image is None:
            return 1
    else:
        print("ERROR: no cloud image specified")
        return 1
    snapshot = qemucfg.create_tempfile_qcow2(image)
    qemucfg.add_disk_qcow2(snapshot.name, bootindex = 1)

    # cloud init iso
    cloud = CloudInitISO(hostname.replace('.','-'))
    cloud.add_user(options.userid, passwd = options.passwd, sudo = True)
    for pkg in options.package:
        cloud.add_package(pkg)
    for f in options.copyfiles:
        cloud.add_file_copy(f)
    for cmd in options.runcmd:
        cloud.add_runcmd(cmd)
    iso = cloud.create_iso()
    qemucfg.add_cdrom_iso(iso.name)

    # go!
    cmdline += qemucfg.cmdline()
    if options.dryrun:
        print("=== cloud image ===")
        print(image)
        print("=== cloud init ===")
        cloud.dump_user()
        print("=== firmware ===")
        print(fw.data)
        print("=== qemu cmdline ===")
        print(cmdline)
    else:
        xterm_title(title)
        subprocess.run(cmdline, check = True)

    # cleanup
    if varsfile:
        varsfile.close()
    snapshot.close()
    iso.close()
    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Emanuele Giuseppe Esposito
#
# pylint: disable=too-many-instance-attributes,consider-iterating-dictionary
""" handle UKI addons """
import os
import sys
import logging
import shutil
import argparse
import hashlib
import subprocess

from virt.firmware.efi import ucs16

from virt.firmware.bootcfg import bootcfg
from virt.firmware.bootcfg import linuxcfg

ESP_PATH = linuxcfg.LinuxOsInfo().esp_path()
DEFAULT_RPM_LOCATION = '/usr/lib/modules/'
DEFAULT_RPM_GLOBAL_LOCATION = DEFAULT_RPM_LOCATION + 'extra.d/'
GLOBAL_ADDONS_LOCATION = ESP_PATH + '/loader/addons/'

UKIFY_PATH = None
for udir in ('/usr/bin', '/usr/lib/systemd'):
    ubin = f'{udir}/ukify'
    if os.path.exists(ubin):
        UKIFY_PATH = ubin


def get_devpath(optdata):
    if optdata:
        if optdata and len(optdata) >= 4 and optdata[0] != 0 and optdata[1] == 0:
            path = ucs16.from_ucs16(optdata, 0)
            if path:
                return ESP_PATH + str(path).split()[0].replace('\\', '/') + '.extra.d/'
    return None


def check_input_addon_exist(addon):
    """
    Check if the addon is an actual addon and exists.
    Returns the full path to the addon.
    """
    if not addon.endswith('.addon.efi'):
        logging.error('addon must end with .addon.efi')
        sys.exit(1)
    if not os.path.exists(addon):
        logging.error('addon %s does not exist', addon)
        sys.exit(1)
    return addon


def check_global_addon_exist(fail=True):
    """
    Check if GLOBAL_ADDONS_LOCATION exist
    """
    if os.path.exists(GLOBAL_ADDONS_LOCATION):
        return True
    if fail:
        logging.error('Global addons location %s does not exist', GLOBAL_ADDONS_LOCATION)
        sys.exit(1)
    return False


def check_global_addon_option(options, fail=True):
    """
    Checks if --global is not given together with --uki-path or --uki-title.
    Returns GLOBAL_ADDONS_LOCATION
    """
    if options.uki_title or options.uki_path:
        if fail:
            logging.error('destination for the addon is either global addons or specific uki')
            sys.exit(1)
        return False
    return GLOBAL_ADDONS_LOCATION


def check_title_option(cfg, options):
    """
    Given a title, finds its entry in cfg and returns the path to the UKI
    """
    if options.uki_path:
        logging.error('cannot specify --title and --dest-uki together')
        sys.exit(1)
    nr = cfg.find_title_entry(options.uki_title)
    if nr is None:
        logging.error('boot entry %s does not exist', options.uki_title)
        sys.exit(1)
    entry = cfg.get_entry(nr)
    if entry is None:
        logging.error('boot entry %s does not exist', options.uki_title)
        sys.exit(1)
    path = get_devpath(entry.optdata)
    if path:
        return path
    logging.error('boot entry %s does not contain a device path', options.uki_title)
    sys.exit(1)


def check_path_option(uki):
    """
    Checks that the given UKI path (or extra.d path) exists.
    Returns the path to the .extra.d folder, but does not check if that actually exists
    """
    if not os.path.exists(uki):
        logging.error('File/folder %s not found', uki)
        sys.exit(1)
    # .efi file, append .extra.d/ and return
    if uki.endswith('.efi'):
        return uki + '.extra.d/'
    # either .extra.d or something else, append '/' if missing
    if not uki.endswith('/'):
        uki += '/'
    # must be .extra.d/, otherwise fail
    if not uki.endswith('.extra.d/'):
        logging.error('%s is not a UKI', uki)
        sys.exit(1)
    return uki


def install_uki_addon(cfg, options):
    """
    cp options.install_addon in a specific destination
    *) with no param, fail
    *) with --global, cp in GLOBAL_ADDONS_LOCATION
    *) with --uki-path/uki-title, cp in
       /boot/efi/EFI/Linux/uki_name.efi.extra.d/options.install_addon
    """
    options.install_addon = check_input_addon_exist(options.install_addon)

    destination = None
    if options.global_addon:
        destination = check_global_addon_option(options)
        if not os.path.exists(ESP_PATH):
            logging.error("%s does not exist", ESP_PATH)
            sys.exit(1)
    elif options.uki_title:
        destination = check_title_option(cfg, options)
    elif options.uki_path:
        destination = check_path_option(options.uki_path)
    else:
        logging.error('please specify a target uki with --title/--dest-uki or --global-addons')
        sys.exit(1)

    os.makedirs(destination, exist_ok=True)
    logging.info('Folder where the addon will be installed: %s', destination)
    new_addon_path = destination + os.path.basename(options.install_addon)
    if os.path.exists(new_addon_path):
        logging.error('addon %s exists already', new_addon_path)
        sys.exit(1)

    shutil.copy(options.install_addon, destination)
    logging.info('Installed addon %s to %s', options.install_addon, destination)


def rm_addon(path, fail=True):
    if os.path.exists(path):
        os.remove(path)
        return True
    if fail:
        logging.error('addon %s does not exist', path)
        sys.exit(1)
    return False


def rm_uki_addon(cfg, options):
    """
    rm options.rm_addon
    *) with no param, just behave like `rm`.
    *) with --global, rm GLOBAL_ADDONS_LOCATION/options.rm_addon
    *) with --uki-path/uki-title, rm /boot/efi/EFI/Linux/uki_name.efi.extra.d/options.rm_addon
    """
    if options.global_addon:
        check_global_addon_exist()
        destination = check_global_addon_option(options) + options.rm_addon
    elif options.uki_title:
        destination = check_title_option(cfg, options) + options.rm_addon
    elif options.uki_path:
        destination = check_path_option(options.uki_path) + options.rm_addon
    else:
        destination = check_input_addon_exist(options.rm_addon)

    rm_addon(destination)
    logging.info('Removed addon %s', destination)


def get_file_sha(path):
    sha256 = hashlib.sha256()

    with open(path, 'rb') as f:
        BUF_SIZE = 65536
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha256.update(data)

    return sha256.hexdigest()


def update_installed_addon(path, options, fail=True):
    if not os.path.exists(path):
        if fail:
            logging.error('%s does not exist', path)
            sys.exit(1)
        return False

    addon = os.path.basename(options.update_addon)
    installed_addon_path = path + addon
    if not os.path.exists(installed_addon_path):
        if fail:
            logging.error('addon %s does not exist', installed_addon_path)
            sys.exit(1)
        return False
    logging.debug('Found addon %s', installed_addon_path)

    old_addon_sha = get_file_sha(installed_addon_path)
    new_addon_sha = get_file_sha(options.update_addon)
    if old_addon_sha != new_addon_sha:
        shutil.copy(options.update_addon, installed_addon_path)
        logging.info('Updated addon %s to %s', options.update_addon, installed_addon_path)
    else:
        logging.info("Addon %s unchanged", installed_addon_path)
    return True


def update_uki_addon(cfg, options):
    """
    cp options.update_addon replacing another one already installed.
    Replacement is only done if name matches and sha256 differs.
    *) with no param, update all *installed* UKIs and GLOBAL_ADDONS_LOCATION if
       there is an addon matching. Replace *all* matching
    *) with --global, just update the one in GLOBAL_ADDONS_LOCATION
    *) with --uki-path/uki-title, update the one in
       /boot/efi/EFI/Linux/uki_name.efi.extra.d/options.update_addon
    """
    options.update_addon = check_input_addon_exist(options.update_addon)

    destination = None
    if options.global_addon:
        destination = check_global_addon_option(options)
        check_global_addon_exist()
        update_installed_addon(destination, options)
    elif options.uki_title:
        destination = check_title_option(cfg, options)
        update_installed_addon(destination, options)
    elif options.uki_path:
        destination = check_path_option(options.uki_path)
        update_installed_addon(destination, options)
    else:
        # update all UKIs
        installed_addon = False
        if check_global_addon_exist(fail=False):
            installed_addon = update_installed_addon(GLOBAL_ADDONS_LOCATION, options, fail=False)
        for entry in cfg.bentr.values():
            destination = get_devpath(entry.optdata)
            if destination:
                installed_addon |= update_installed_addon(destination, options, fail=False)
        if not installed_addon:
            logging.error('no addon found to update in global addons folder or installed ukis')
            sys.exit(1)


def print_addons_folder(path, options):
    found = 0
    if not os.path.exists(path):
        logging.error('%s does not exist!', path)
        return -1

    for addon in os.listdir(path):
        if not addon.endswith('.addon.efi'):
            continue
        fullpath = path + addon
        print(f'  {fullpath}')
        found += 1
        if options.verbose:
            cmd = f'{UKIFY_PATH} inspect {fullpath}'.split()
            out = subprocess.check_output(cmd, text=True)
            out = '    ' + out.replace('\n', '\n    ')
            print(out, end=None)
    return found


def show_all_addons(destination, options, descr=None, global_addons=True):
    print('#' * len(destination))
    descr = f'({descr})' if descr else ""
    print(f'{destination} {descr}')
    print('#' * len(destination))

    found = 0
    if global_addons and check_global_addon_exist(fail=False):
        found += print_addons_folder(GLOBAL_ADDONS_LOCATION, options)
    found += print_addons_folder(destination, options)
    if found == 0:
        print('    No UKI addons found\n\n')
    elif not options.verbose:
        print()


def list_addons_path(options):
    """
    list all addons in a folder
    *) with no params, look into DEFAULT_RPM_GLOBAL_LOCATION and
       DEFAULT_RPM_LOCATION/*/*.efi.extra.d/
    *) with --verbose, print all addon sections
    """
    #options.list_addons is False if undefined, is True if defined without
    # folder and a string if the folder is provided
    destination = options.list_addons
    if options.list_addons is True:
        # look in DEFAULT_RPM_GLOBAL_LOCATION
        show_all_addons(DEFAULT_RPM_GLOBAL_LOCATION, options,
                        descr="global addons in root fs", global_addons=False)
        # look into DEFAULT_RPM_LOCATION/<uname -r>/<uki>.efi.extra.d
        if not os.path.exists(DEFAULT_RPM_LOCATION):
            logging.error('%s does not exist!', DEFAULT_RPM_LOCATION)
            return
        found = 0
        for kernel in os.listdir(DEFAULT_RPM_LOCATION):
            kernel_path = DEFAULT_RPM_LOCATION + kernel
            for uki in os.listdir(kernel_path):
                # /lib/uname-r/<UKI>.efi.extra.d is a legacy location which is incompatible
                # with modern kernel-install, see https://github.com/systemd/systemd/issues/39569
                if not uki.endswith('.efi.extra.d') and not uki.endswith('.efi.extras.optional'):
                    continue
                found += 1
                uki_path = kernel_path + '/' + uki
                show_all_addons(uki_path, options, descr="uki-specific addons in root fs",
                                global_addons=False)
        if found == 0:
            print(f'No UKI addons found in {DEFAULT_RPM_LOCATION}*/*.efi.extras.optional/')
    else:
        show_all_addons(destination, options, global_addons=False)


def show_installed_addons(cfg, options):
    """
    list all addons installed
    *) with no params, list all addons in GLOBAL_ADDONS_LOCATION and all *installed* UKIs.
    *) with --global, show only the global addons. If GLOBAL_ADDONS_LOCATION does not exist, fail
    *) with --uki-path/uki-title, show all addons used by a specific UKI
    *) with --verbose, print all addon sections
    """
    destination = None
    descr = None
    if options.global_addon:
        destination = check_global_addon_option(options, fail=options.global_addon)
        check_global_addon_exist()
        descr = "global addons in ESP"
    elif options.uki_title:
        destination = check_title_option(cfg, options)
    elif options.uki_path:
        destination = check_path_option(options.uki_path)

    if descr is None:
        descr = "uki-specific addons in ESP"

    if destination:
        show_all_addons(destination, options, descr=descr,
                        global_addons=options.global_addon is None)
    else:
        for entry in cfg.bentr.values():
            path = get_devpath(entry.optdata)
            if path:
                show_all_addons(path, options, descr=descr)


# pylint: disable=too-many-boolean-expressions,too-many-branches,too-many-statements
def main():
    parser = argparse.ArgumentParser(
        description = 'show and manage UKI addons')

    parser.add_argument('-l', '--loglevel', dest = 'loglevel', type = str, default = 'info',
                        help = 'set loglevel to LEVEL', metavar = 'LEVEL')
    parser.add_argument('--vars', dest = 'varsfile', type = str,
                        help = 'read edk2 vars from FILE', metavar = 'FILE')
    parser.add_argument('-v', '--verbose', dest = 'verbose',
                        action = 'store_true', default = False,
                        help = 'print more details')

    group = parser.add_argument_group('update unified kernel image (UKI) addons')
    group.add_argument('--install-addon', dest = 'install_addon', type = str,
                       help = 'install addon FILE', metavar = 'FILE')
    group.add_argument('--update-addon', dest = 'update_addon', type = str,
                       help = 'update addon FILE', metavar = 'FILE')
    group.add_argument('--remove-addon', dest = 'rm_addon', type = str,
                       help = 'remove addon FILE', metavar = 'FILE')
    group.add_argument('--list-local-addons', dest = 'list_addons', type = str, nargs='?',
                       const = True,
                       help = 'list all addons in PATH (leave it empty to show addons'
                              ' installed in root fs)',
                       metavar = 'PATH')
    parser.add_argument('--show-installed', dest = 'show',
                        action = 'store_true', default = False,
                        help = 'show all installed addons in the ESP')

    group = parser.add_argument_group('options for UKI addons updates')
    group.add_argument('--global', dest = 'global_addon',
                       action = 'store_true', default = False,
                       help = f'work only with addons in  {GLOBAL_ADDONS_LOCATION}')
    group.add_argument('--uki-title', dest = 'uki_title', type = str,
                       help = 'work only with addons in UKI labelled TITLE',
                       metavar = 'TITLE')
    group.add_argument('--uki-path', dest = 'uki_path', type = str,
                       help = 'work only with addons used by the UKI in PATH',
                       metavar = 'PATH')

    options = parser.parse_args()

    logging.basicConfig(format = '%(levelname)s: %(message)s',
                        level = getattr(logging, options.loglevel.upper()))

    # read info
    if options.varsfile:
        cfg = bootcfg.VarStoreEfiBootConfig(options.varsfile)
    else:
        cfg = linuxcfg.LinuxEfiBootConfig()

    # apply updates
    if options.install_addon:
        install_uki_addon(cfg, options)
    elif options.rm_addon:
        rm_uki_addon(cfg, options)
    elif options.update_addon:
        update_uki_addon(cfg, options)
    elif options.list_addons:
        list_addons_path(options)
    else:
        show_installed_addons(cfg, options)

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2026 Gerd Hoffmann
#
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
""" generate cloud init iso images """
import os
import sys
import shutil
import pathlib
import tempfile
import subprocess

import yaml

try:
    import crypt_r as crypt
except ImportError:
    import crypt


class LiteralStr(str):
    """Marker class: dump as literal block scalar (|)."""

def literal_representer(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")

class CloudInitDumper(yaml.SafeDumper):
    """Safe YAML dumper with support for LiteralStr literal block scalars."""

CloudInitDumper.add_representer(LiteralStr, literal_representer)


class CloudInitISO:
    """ cloud init iso image generator """

    def __init__(self, hostname):
        self.config = {
            "hostname": hostname,
        }
        self.meta = {
            "local-hostname": hostname
        }

    @staticmethod
    def crypt(passwd):
        return crypt.crypt(passwd, crypt.mksalt(method=crypt.METHOD_SHA512))

    def add_user(self, name, passwd = None, ssh_keys = None, sudo = False, shell = "/bin/bash"):
        user = {
            "name"  : name,
            "shell" : shell
        }
        if passwd:
            user["passwd"] = self.crypt(passwd)
            user["lock_passwd"] = False
        if ssh_keys:
            user["ssh_authorized_keys"] = ssh_keys
        if sudo:
            user["sudo"] = ["ALL=(ALL) NOPASSWD:ALL"]
        if not self.config.get("users"):
            self.config["users"] = []
        self.config["users"].append(user)

    def add_package(self, package_name):
        """ package to install """
        if not self.config.get("packages"):
            self.config["packages"] = []
        self.config["packages"].append(package_name)

    def add_runcmd(self, command):
        """ command to run on first boot """
        if not self.config.get("runcmd"):
            self.config["runcmd"] = []
        self.config["runcmd"].append(command)

    def add_bootcmd(self, command):
        """ command to run on every boot """
        if not self.config.get("bootcmd"):
            self.config["bootcmd"] = []
        self.config["bootcmd"].append(command)

    def add_file_content(self, filename, content, owner = 'root:root'):
        """ file to write to disk """
        if not self.config.get("write_files"):
            self.config["write_files"] = []
        wf = {
            "path"    : filename,
            "owner"   : owner,
            "content" : content,
        }
        self.config["write_files"].append(wf)

    def add_file_copy(self, filename):
        try:
            with open(filename, 'r', encoding = 'utf-8') as f:
                content = f.read()
            self.add_file_content(filename, LiteralStr(content))
        except UnicodeDecodeError:
            with open(filename, 'rb') as f:
                content = f.read()
            self.add_file_content(filename, content)

    # pylint: disable=consider-using-with
    def create_iso(self):
        iso = tempfile.NamedTemporaryFile(suffix = '.iso')
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = pathlib.Path(tmpdir)

            user_data_file = tmpdir / "user-data"
            with user_data_file.open("w") as f:
                f.write("#cloud-config\n")
                yaml.dump(self.config, f, default_flow_style=False,
                          sort_keys=False, Dumper=CloudInitDumper)

            meta_data_file = tmpdir / "meta-data"
            with meta_data_file.open("w") as f:
                yaml.dump(self.meta, f, default_flow_style=False,
                          sort_keys=False, Dumper=CloudInitDumper)

            tmtdata = os.environ.get('TMT_TEST_DATA')
            if tmtdata:
                dstdir = pathlib.Path(tmtdata) / "cloud-init"
                os.makedirs(dstdir, exist_ok = True)
                shutil.copy(user_data_file, dstdir / "user-data")
                shutil.copy(meta_data_file, dstdir / "meta-data")

            subprocess.run([
                "xorrisofs",
                "-output", iso.name,
                "-volid", "cidata",
                "-joliet",
                "-rock",
                str(user_data_file),
                str(meta_data_file)
            ], capture_output = True, check=True)

        return iso

    def dump_user(self):
        y = yaml.dump(self.config, default_flow_style=False,
                      sort_keys=False, Dumper=CloudInitDumper)
        print(y, end='')

def main():
    c = CloudInitISO("test-host")
    c.add_user("user", "passwd", sudo = True)
    c.add_package("pciutils")
    c.add_bootcmd("echo hello bootcmd")
    c.add_runcmd("echo hello runcmd #1")
    c.add_runcmd("echo hello runcmd #2")
    c.add_file_content("/root/test", "hello world")
    c.add_file_copy("/etc/resolv.conf")
    c.dump_user()

if __name__ == '__main__':
    sys.exit(main())



""" alpine linux download """
import os
import yaml
import requests

from virt.qemutest.distro.core import FetchDistroCore

class FetchAlpineIso(FetchDistroCore):
    """ class for alpine linux iso download """

    mirror = 'https://dl-cdn.alpinelinux.org/alpine/latest-stable/releases/'

    def __init__(self, mirror = None, arch = None):
        super().__init__(arch)
        if mirror:
            self.mirror = mirror

    def baseurl(self):
        return f'{self.mirror}{self.arch}/'

    def image_id(self, flavor):
        return f'{flavor}-{self.arch}-liveiso'

    def find_image(self, flavor):
        req = requests.get(f'{self.baseurl()}latest-releases.yaml', timeout = 60)
        req.raise_for_status()
        entries = yaml.load(req.text, Loader=yaml.SafeLoader)
        for e in entries:
            if e.get('flavor') != flavor:
                continue
            filename = e.get('file')
            checksum = e.get('sha256')
            return (filename, checksum)
        raise RuntimeError(f'alpine flavor {flavor} not found')

    def fetch_image(self, flavor = 'alpine-virt', imgdir = 'distro'):
        print(f'# ----- {flavor} -----')
        (latest, checksum) = self.find_image(flavor)
        url = f'{self.baseurl()}{latest}'
        out = f'{imgdir}/{latest}'
        os.makedirs(imgdir, exist_ok = True)
        self.fetch_image_checksum(url, out, checksum)
        self.index_update(imgdir, self.image_id(flavor), latest, checksum)

    def iso_url(self, flavor = 'alpine-virt'):
        (latest, checksum) = self.find_image(flavor)
        url = f'{self.baseurl()}{latest}'
        return url



""" centos stream download """
import os

from virt.qemutest.distro.core import FetchDistroCore, FetchNetinst

class FetchCentosCloud(FetchDistroCore):
    """ class for centos stream cloud image download """

    centos = "https://cloud.centos.org/centos/"

    def __init__(self, mirror = None, arch = None):
        super().__init__(arch)
        self.mirror = mirror
        if mirror is None:
            self.mirror = self.centos

    def baseurl(self, release):
        return f'{self.mirror}{release}-stream/{self.arch}/images/'

    def image_id(self, release):
        return f'centos-stream-{release}-{self.arch}-cloud'

    def find_image(self, release, pattern):
        checksum = f'{self.baseurl(release)}CHECKSUM'
        return self.checksum_find_latest(checksum, pattern)

    def fetch_image(self, release = 10, pattern = 'GenericCloud', imgdir = 'distro'):
        print(f'# ----- centos {release} ({pattern}) -----')
        (latest, checksum) = self.find_image(release, pattern)
        url = f'{self.baseurl(release)}{latest}'
        out = f'{imgdir}/{latest}'
        os.makedirs(imgdir, exist_ok = True)
        self.fetch_image_checksum(url, out, checksum)
        self.index_update(imgdir, self.image_id(release), latest, checksum)

class FetchCentosNetinst(FetchNetinst):
    """ class for centos stream netinst iso download """

    centos = "https://mirror.stream.centos.org/"

    def __init__(self, mirror = None, arch = None):
        if mirror is None:
            mirror = self.centos
        super().__init__(mirror, arch)

    def baseurl(self, release):
        repo = f'{self.mirror}{release}-stream/BaseOS/{self.arch}/os/'
        return repo



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2026 Gerd Hoffmann
#
""" build qemu command line """
import os
import json
import tempfile
import subprocess

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-instance-attributes
class QemuCfg():
    """ build qemu command line """

    # pylint: disable=fixme
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def __init__(self, logger = None, pcie = True, scsi = True, nvme = False, sata = False):

        # config
        self.logger = logger
        self.use_pcie = pcie
        self.use_scsi = scsi
        self.use_nvme = nvme
        self.use_sata = sata

        # TODO: autodetect cbit
        self.sev_cbit = 51
        self.sev_kernel_hashes = False

        # state
        self.have_scsi = False
        self.have_xhci = False
        self.have_bot = False
        self.have_virtio_serial = False
        self.have_nvme = False

        # generate ids
        self.ids = {}
        self.pcie = 0

        # device / object lists
        self.blkdev = []
        self.netdev = []
        self.chardev = []
        self.devices = []
        self.objects = []
        self.machine = []


    def get_id(self, kind):
        self.ids.setdefault(kind, 0)
        retval = f'{kind}{self.ids[kind]}'
        self.ids[kind] += 1
        return retval

    @staticmethod
    def mkarg(value):
        if isinstance(value, dict):
            return json.dumps(value)
        return value

    def cmdline(self):
        cmdline = []
        for obj in self.objects:
            cmdline.append('-object')
            cmdline.append(self.mkarg(obj))
        for blk in self.blkdev:
            cmdline.append('-blockdev')
            cmdline.append(self.mkarg(blk))
        for net in self.netdev:
            cmdline.append('-netdev')
            cmdline.append(self.mkarg(net))
        for char in self.chardev:
            cmdline.append('-chardev')
            cmdline.append(self.mkarg(char))
        for dev in self.devices:
            cmdline.append('-device')
            cmdline.append(self.mkarg(dev))
        for m in self.machine:
            cmdline.append('-machine')
            cmdline.append(m)
        return cmdline

    # --- devices ---

    def add_pcie_port(self):
        name = f'port{self.pcie}'
        slot = 8 + (self.pcie >> 3)
        func = self.pcie & 7
        device = { 'driver'  : 'pcie-root-port',
                   'id'      : name,
                   'bus'     : 'pcie.0',
                   'addr'    : f'{slot:x}.{func:x}',
                   'chassis' : self.pcie }
        if func == 0:
            device['multifunction'] = True
        self.devices.append(device)
        self.pcie += 1
        return name

    def add_virtio_device(self, kind, args = None):
        devname = f'virtio-{kind}-pci'
        port = None
        if self.use_pcie:
            port = self.add_pcie_port()

        if args is None:
            args = {}
        if isinstance(args, dict):
            # json
            device = dict(args)
            device['driver'] = devname
            if port:
                device['bus'] = port
        else:
            # string
            device = f'{devname},{args}'
            if port:
                device += f',bus={port}'

        self.devices.append(device)

    def add_virtio_net(self, netdev, bootindex = None):
        args = { 'netdev' : netdev }
        if bootindex is not None:
            args['bootindex'] = bootindex
        self.add_virtio_device('net', args)

    def add_virtio_blk(self, blkdev, bootindex = None):
        args = { 'drive' : blkdev }
        if bootindex is not None:
            args['bootindex'] = bootindex
        self.add_virtio_device('blk', args)

    def add_virtio_scsi(self):
        self.add_virtio_device('scsi', { 'id' : 'virtio' })
        self.have_scsi = True

    def add_virtio_rng(self):
        self.add_virtio_device('rng')

    def add_virtio_serial(self):
        self.add_virtio_device('serial')
        self.have_virtio_serial = True

    def add_virtio_serial_port(self, chardev, name):
        if not self.have_virtio_serial:
            self.add_virtio_serial()
        device = { 'driver'  : 'virtserialport',
                   'chardev' : chardev,
                   'name'    : name }
        self.devices.append(device)

    def add_scsi_hd(self, blkdev, bootindex = None):
        if not self.have_scsi:
            self.add_virtio_scsi()
        device = { 'driver' : 'scsi-hd',
                   'bus'    : 'virtio.0',
                   'drive'  : blkdev }
        if bootindex is not None:
            device['bootindex'] = bootindex
        self.devices.append(device)

    def add_scsi_cd(self, blkdev, bootindex = None):
        if not self.have_scsi:
            self.add_virtio_scsi()
        device = { 'driver' : 'scsi-cd',
                   'bus'    : 'virtio.0',
                   'drive'  : blkdev }
        if bootindex is not None:
            device['bootindex'] = bootindex
        self.devices.append(device)

    def add_sata_hd(self, blkdev, bootindex = None):
        device = { 'driver' : 'ide-hd',
                   'bus'    : self.get_id('ide.'),
                   'drive'  : blkdev }
        if bootindex is not None:
            device['bootindex'] = bootindex
        self.devices.append(device)

    def add_sata_cd(self, blkdev, bootindex = None):
        device = { 'driver' : 'ide-cd',
                   'bus'    : self.get_id('ide.'),
                   'drive'  : blkdev }
        if bootindex is not None:
            device['bootindex'] = bootindex
        self.devices.append(device)

    def add_nvme_ctrl(self):
        device = { 'driver' : 'nvme',
                   'serial' : '1234' }
        if self.use_pcie:
            port = self.add_pcie_port()
            device['bus'] = port
        self.devices.append(device)
        self.have_nvme = True

    def add_nvme_ns(self, blkdev, bootindex = None):
        if not self.have_nvme:
            self.add_nvme_ctrl()
        device = { 'driver' : 'nvme-ns',
                   'drive'  : blkdev }
        if bootindex is not None:
            device['bootindex'] = bootindex
        self.devices.append(device)

    def add_usb_xhci(self):
        ports = 8
        device = { 'driver' : 'qemu-xhci',
                   'p2'     : ports,
                   'p3'     : ports }
        if self.use_pcie:
            port = self.add_pcie_port()
            device['bus'] = port
        self.devices.append(device)
        self.have_xhci = True

    def add_usb_cd(self, blkdev, bootindex = None):
        if not self.have_xhci:
            self.add_usb_xhci()
        if not self.have_bot:
            self.devices.append({ 'driver' : 'usb-bot',
                                  'id'     : 'usb.bot' })
            self.have_bot = True
        device = { 'driver' : 'scsi-cd',
                   'bus'    : 'usb.bot.0',
                   'drive'  : blkdev }
        if bootindex is not None:
            device['bootindex'] = bootindex
        self.devices.append(device)

    def add_isa_debugcon(self, chardev, iobase = 0x402):
        device = { 'driver'  : 'isa-debugcon',
                   'chardev' : chardev,
                   'iobase'  : iobase }
        self.devices.append(device)

    # --- network ---

    def add_netdev_user(self, nid):
        netdev = { 'type' : 'user',
                   'id'   : nid }
        self.netdev.append(netdev)

    # --- storage ---

    def add_blkdev_qcow2(self, bid, filename):
        if self.logger:
            self.logger.info('# blkdev/qcow2: id=%s file=%s', bid, filename)
        blkdev = { 'node-name' : bid,
                   'driver'    : 'qcow2',
                   'file'      : { 'driver'   : 'file',
                                   'filename' : filename }}
        self.blkdev.append(blkdev)

    def add_blkdev_iso(self, bid, filename):
        if self.logger:
            self.logger.info('# blkdev/iso: id=%s file=%s', bid, filename)
        blkdev = { 'node-name' : bid,
                   'driver'    : 'file',
                   'filename'  : filename,
                   'read-only' : True }
        self.blkdev.append(blkdev)

    # --- chardev ---

    def add_chardev_logfile(self, cid, filename):
        chardev = f'file,id={cid},path={filename}'
        self.chardev.append(chardev)

    # --- helpers ---

    # pylint: disable=consider-using-with
    def create_tempfile_qcow2(self, image):
        copy = tempfile.NamedTemporaryFile(suffix = '.qcow2')
        if self.logger:
            self.logger.info('# qcow2: copy=%s backing=%s', copy.name, image)
        qemuimg = [ 'qemu-img', 'create', '-q', '-f', 'qcow2', '-F', 'qcow2',
                    '-b', os.path.abspath(image), copy.name ]
        subprocess.run(qemuimg, check = True)
        return copy

    def create_empty_qcow2(self, filename, size = '8G'):
        if os.path.exists(filename):
            if self.logger:
                self.logger.info("# qcow2: skip=%s [file exists]", filename)
            return

        if self.logger:
            self.logger.info("# qcow2: create=%s size=%s", filename, size)
        qemuimg = [ 'qemu-img', 'create', '-q', '-f', 'qcow2', filename, size ]
        subprocess.run(qemuimg, check = True)

    def add_disk_qcow2(self, filename, bootindex = None):
        bid = self.get_id('blk')
        self.add_blkdev_qcow2(bid, filename)
        if self.use_nvme:
            self.add_nvme_ns(blkdev = bid, bootindex = bootindex)
        elif self.use_scsi:
            self.add_scsi_hd(blkdev = bid, bootindex = bootindex)
        elif self.use_sata:
            self.add_sata_hd(blkdev = bid, bootindex = bootindex)
        else:
            self.add_virtio_blk(blkdev = bid, bootindex = bootindex)

    def add_cdrom_iso(self, filename, bootindex = None):
        bid = self.get_id('blk')
        self.add_blkdev_iso(bid, filename)
        if self.use_scsi:
            self.add_scsi_cd(blkdev = bid, bootindex = bootindex)
        elif self.use_sata:
            self.add_sata_cd(blkdev = bid, bootindex = bootindex)
        else:
            self.add_usb_cd(blkdev = bid, bootindex = bootindex)

    def add_network(self, bootindex = None):
        nid = self.get_id('net')
        self.add_netdev_user(nid = nid)
        self.add_virtio_net(netdev = nid, bootindex = bootindex)

    def add_virtio_serial_log(self, name, filename):
        cid = self.get_id('char')
        self.add_chardev_logfile(cid, filename)
        self.add_virtio_serial_port(cid, name)

    def add_x64_firmware_log(self, filename):
        cid = self.get_id('char')
        self.add_chardev_logfile(cid, filename)
        self.add_isa_debugcon(cid)

    # --- confidential ---

    def set_sev_cbit(self, cbit):
        self.sev_cbit = cbit

    def set_sev_kernel_hashes(self, kernel_hashes):
        self.sev_kernel_hashes = kernel_hashes

    def add_confidential_tdx(self):
        oid = self.get_id('obj')
        tdx = { 'qom-type' : 'tdx-guest',
                'id'       : oid }
        if os.path.exists('/run/tdx-qgs/qgs.socket'):
            qgs = { 'type' : 'unix',
                    'path' : '/run/tdx-qgs/qgs.socket' }
            tdx['quote-generation-socket'] = qgs
        self.objects.append(tdx)
        self.machine.append(f'confidential-guest-support={oid}')

    def add_confidential_sev_common(self, sev):
        oid = self.get_id('obj')
        sev['id']       = oid
        sev['cbitpos']  = self.sev_cbit
        sev['reduced-phys-bits'] = 1
        sev['kernel-hashes'] = self.sev_kernel_hashes
        self.objects.append(sev)
        self.machine.append(f'confidential-guest-support={oid}')

    def add_confidential_sev(self):
        sev = { 'qom-type' : 'sev-guest' }
        self.add_confidential_sev_common(sev)

    def add_confidential_sev_es(self):
        sev = { 'qom-type' : 'sev-guest',
                'policy'   : 5}
        self.add_confidential_sev_common(sev)

    def add_confidential_snp(self):
        sev = {'qom-type' : 'sev-snp-guest'}
        self.add_confidential_sev_common(sev)

    def add_confidential(self, cc_mode):
        if cc_mode == 'sev':
            self.add_confidential_sev()
        elif cc_mode in [ 'sev-es', 'es' ]:
            self.add_confidential_sev_es()
        elif cc_mode in [ 'sev-snp', 'snp' ]:
            self.add_confidential_snp()
        elif cc_mode == 'tdx':
            self.add_confidential_tdx()



""" fedora download """
import os

from virt.qemutest.distro.core import FetchDistroCore, FetchNetinst

class FetchFedoraCloud(FetchDistroCore):
    """ class for fedora cloud image download """

    fedora = "https://download.fedoraproject.org/pub"

    def __init__(self, mirror = None, arch = None):
        super().__init__(arch)
        self.mirror = mirror
        if mirror is None:
            self.mirror = self.fedora

    def baseurl(self, release):
        subdir = 'releases'
        location = f'fedora/linux/{subdir}/{release}/Cloud/{self.arch}/images'
        return f'{self.mirror}/{location}/'

    def image_id(self, release):
        return f'fedora-{release}-{self.arch}-cloud'

    def find_image(self, release, pattern):
        index = self.baseurl(release)
        checksum_name = self.indexhtml_find_file(index, 'CHECKSUM')
        if checksum_name is None:
            return (None, None)
        print(f'# checksum file: {checksum_name}')
        checksum_url = f'{self.baseurl(release)}{checksum_name}'
        return self.checksum_find_latest(checksum_url, pattern)

    def fetch_image(self, release = 44, pattern = 'Generic', imgdir = 'distro'):
        print(f'# ----- fedora {release} ({pattern}) -----')
        (latest, checksum) = self.find_image(release, pattern)
        url = f'{self.baseurl(release)}{latest}'
        out = f'{imgdir}/{latest}'
        os.makedirs(imgdir, exist_ok = True)
        self.fetch_image_checksum(url, out, checksum)
        self.index_update(imgdir, self.image_id(release), latest, checksum)

class FetchFedoraUKICloud(FetchFedoraCloud):
    """ class for fedora uki cloud image download """

    def image_id(self, release):
        return f'fedora-uki-{release}-{self.arch}-cloud'

    def fetch_image(self, release = 44, pattern = 'UKI', imgdir = 'distro'):
        super().fetch_image(release, pattern, imgdir)

class FetchFedoraNetinst(FetchNetinst):
    """ class for fedora server netinst iso download """

    fedora = "https://download.fedoraproject.org/pub"

    def __init__(self, mirror = None, arch = None):
        if mirror is None:
            mirror = self.fedora
        super().__init__(mirror, arch)

    def baseurl(self, release):
        if release == 'rawhide':
            subdir = 'development'
        else:
            subdir = 'releases'
        location = f'fedora/linux/{subdir}/{release}/Everything/{self.arch}/os'
        return f'{self.mirror}/{location}/'



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
""" print and edit efi sigdb files """
import sys
import argparse

from cryptography.hazmat.primitives import serialization

from virt.firmware.efi import guids
from virt.firmware.efi import efivar
from virt.firmware.efi import siglist
from virt.peutils import pesign

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', dest = 'input', type = str,
                        help = 'read efi sigdb FILE', metavar = 'FILE')
    parser.add_argument('-o', '--output', dest = 'output', type = str,
                        help = 'write efi sigdb FILE.', metavar = 'FILE')
    parser.add_argument('--add-cert', dest = 'certs',  action = 'append', nargs = 2,
                        help = 'add x509 cert to sigdb, loaded in pem format ' +
                        'from FILE and with owner GUID, can be specified multiple times',
                        metavar = ('GUID', 'FILE'))
    parser.add_argument('--add-hash', dest = 'hashes',  action = 'append', nargs = 2,
                        help = 'add sha256 hash to sigdb, with owner GUID, '
                        'can be specified multiple times',
                        metavar = ('GUID', 'HASH'))
    parser.add_argument('-x', '--extract', dest = 'extract',
                        action = 'store_true', default = False,
                        help = 'extract certificates')
    parser.add_argument('-p', '--print', dest = 'print',
                        action = 'store_true', default = False,
                        help = 'print sigdb')
    options = parser.parse_args()

    if options.input:
        with open(options.input, "rb") as f:
            sigdb = siglist.EfiSigDB(f.read())
    else:
        sigdb = siglist.EfiSigDB()

    if options.certs:
        for item in options.certs:
            sigdb.add_cert(guids.parse_str(item[0]), item[1])

    if options.hashes:
        for item in options.hashes:
            sigdb.add_hash(guids.parse_str(item[0]),
                           bytes.fromhex(item[1]))

    if options.print and sigdb:
        for slist in sigdb:
            efivar.EfiVarList.print_siglist(slist)
            if options.extract and slist.x509:
                scn = pesign.cert_common_name(slist.x509.subject)
                fn = "".join(x for x in scn if x.isalnum()) + '.pem'
                print(f'    >>> {fn}')
                with open(fn, 'wb') as f:
                    f.write(slist.x509.public_bytes(serialization.Encoding.PEM))

    if options.output:
        with open(options.output, "wb") as f:
            f.write(bytes(sigdb))

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2026 Gerd Hoffmann
#
""" misc utility functions """
import re
import datetime

zapterm1 = re.compile(r'\x1b\[[0-9;:=?!]*[HJhlmnpt]')
zapterm2 = re.compile(r'\x1b\][0-9]*')
zapterm3 = re.compile(r'[\x07\r]')

def zap_term(line):
    """ remove terminal control sequences """
    line = zapterm1.sub('', line)
    line = zapterm2.sub('', line)
    line = zapterm3.sub('', line)
    return line

def iso_now():
    """ return current time in iso format """
    tz = datetime.timezone.utc
    ts = datetime.datetime.now(tz)
    return ts.isoformat()



""" base class for image download """
import os
import re
import json
import time
import hashlib
import datetime
import configparser

import requests

def default_imgdir():
    """ return default distro image directory """
    imgdir = os.environ.get('VIRT_DISTRO_IMGDIR')
    if imgdir is not None:
        return imgdir
    home = os.environ.get('HOME')
    if home is not None:
        return f'{home}/.cache/virt-distro'
    return 'distro'

class DistroList:
    """ list of downloaded distro images """

    def __init__(self, imgdir):
        self.imgdir = imgdir
        self.filename = f'{self.imgdir}/index.json'
        self.index = {}
        self.index_load()

    def index_load(self):
        if not os.path.exists(self.filename):
            return
        with open(self.filename, 'r', encoding = 'utf-8') as f:
            self.index = json.load(f)

    def index_dump(self):
        with open(self.filename, 'w', encoding = 'utf-8') as f:
            json.dump(self.index, f)

    def checksum_dump(self, hashalgo = 'sha256'):
        outfile = f'{self.imgdir}/checksums.{hashalgo}'
        with open(outfile, 'w', encoding = 'utf-8') as f:
            for name in self.index.keys():
                if self.index.get(name).get('hashalgo') != hashalgo:
                    continue
                filename = self.index.get(name).get('filename')
                checksum = self.index.get(name).get('checksum')
                print(f'{checksum} {filename}', file = f)

    @staticmethod
    def mkid(name_ver, kind, arch = None):
        if arch is None:
            arch = os.uname().machine
        return f'{name_ver}-{arch}-{kind}'

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def update(self, image_id, filename,
               checksum, hashalgo = 'sha256',
               url = None):
        self.index[image_id] = { 'filename' : filename,
                                 'checksum' : checksum,
                                 'hashalgo' : hashalgo }
        if url:
            self.index[image_id]['url'] = url
        self.index_dump()
        self.checksum_dump(hashalgo)

    def find_filename(self, image_id):
        entry = self.index.get(image_id)
        if entry is None:
            return None
        filename = entry.get('filename')
        if filename is None:
            return None
        return f'{self.imgdir}/{filename}'


class FetchDistroCore:
    """ base class for image download """

    arch = os.uname().machine
    retries = 5
    timeout = 20

    def __init__(self, arch = None):
        if arch is not None:
            self.arch = arch

    @staticmethod
    def index_load(imgdir):
        dl = DistroList(imgdir)
        return dl.index

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    @staticmethod
    def index_update(imgdir, image_id, filename,
                     checksum, hashalgo = 'sha256',
                     url = None):
        dl = DistroList(imgdir)
        dl.update(image_id, filename,
                  checksum, hashalgo,
                  url = url)

    @staticmethod
    def index_find(imgdir, image_id):
        dl = DistroList(imgdir)
        return dl.find_filename(image_id)

    @staticmethod
    def request(url, stream, start = None, end = None):
        hdr = {}
        if start and end:
            hdr['range'] = f'bytes={start}-{end-1}'
        r = FetchDistroCore.retries
        for i in range(r):
            try:
                req = requests.get(url, headers = hdr,
                                   stream = stream,
                                   timeout = FetchDistroCore.timeout)
            except requests.exceptions.RequestException as e:
                print(f'# request connect error [{i}/{r}]: {e}')
                print(f'# retry in {FetchDistroCore.timeout} seconds')
                time.sleep(FetchDistroCore.timeout)
                continue
            req.raise_for_status()
            if i > 0:
                print(f'# ... connected [{i}/{r}]')
            return req
        return None

    @staticmethod
    def fetch_image_checksum(url, out, checksum):
        """ download one file, verify checksum """

        print(f'# latest: {out}')
        print(f'# sha256: {checksum}')
        if os.path.exists(out):
            print('# skip download, file exists')
            return
        print('# downloading file ...')

        req = FetchDistroCore.request(url, True)
        if req is None:
            raise RuntimeError('connect failed')
        count = 0
        length = int(req.headers.get("content-length", 0))
        lengthm = int(length / (1024*1024))
        hasher = hashlib.sha256()

        with open(out, 'wb') as fh:
            i = req.iter_content(chunk_size = 1024 * 64)
            while True:
                try:
                    chunk = next(i)
                except StopIteration:
                    chunk = None
                except requests.exceptions.RequestException as e:
                    print('')
                    print(f'# request read error: {e}')
                    print('# try reconnect ...')
                    req = FetchDistroCore.request(url, True, count, length)
                    if req is None:
                        os.remove(out)
                        raise RuntimeError('re-connect failed') from e
                    i = req.iter_content(chunk_size = 1024 * 64)
                    continue

                if chunk is None:
                    break
                fh.write(chunk)
                hasher.update(chunk)
                count += len(chunk)
                countm = int(count / (1024*1024))
                if length == 0:
                    print(f'# {countm:4}M', end='\r')
                else:
                    percent = int(count * 100 / length)
                    print(f'# {countm:4}M / {lengthm:4}M - {percent:2}%', end='\r')

        print()
        if hasher.hexdigest() != checksum:
            print(f'# sha256: {hasher.hexdigest()} (mismatch)')
            os.remove(out)
            raise RuntimeError('sha256 mismatch')

        print('# sha256: check ok')

    @staticmethod
    def checksum_find_latest(url, pattern, checksum = 'SHA256'):
        """ find file (and sha256) by downloading CHECKSUM file """

        req = FetchDistroCore.request(url, False)
        if req is None:
            raise RuntimeError('request failed')
        matches = []
        for line in req.text.splitlines():
            if not line.startswith(checksum):
                continue
            if not pattern in line:
                continue
            if 'latest' in line:
                continue
            matches.append(line)
        if len(matches) == 0:
            return (None, None)
        latest = sorted(matches, reverse = True)[0]
        regex = checksum + r' \((.+)\) = ([a-fA-F0-9]{64})'
        m = re.match(regex, latest)
        filename = m.group(1)
        checksum = m.group(2)
        return (filename, checksum)

    @staticmethod
    def indexhtml_find_file(url, pattern):
        """ find file by searching index.html """

        req = FetchDistroCore.request(url, False)
        if req is None:
            raise RuntimeError('request failed')
        regex = rf'href="([^"]*{re.escape(pattern)}[^"]*)"'
        found = re.search(regex, req.text)
        if not found:
            return None
        # some mirrors return absolute not relative links
        elems = found.group(1).split('/')
        return elems[-1]

    @staticmethod
    def find_mirror(url):
        """ find location by following mirror network redirects """
        retries = 5
        while retries > 0:
            print(f'# {url}')
            r = requests.head(url, allow_redirects = False,
                              timeout = FetchDistroCore.timeout)
            if r.status_code == 200:
                print('# -> OK')
                return url
            if r.status_code in (301, 302):
                print(f'# -> Redirect [{r.status_code}]')
                url = r.headers['location']
            else:
                print(f'# -> Oops [{r.status_code}]')
                return None
            retries -= 1
        print('# too many redirects')
        return None

class TreeInfo:
    """ handle .treeinfo files """

    def __init__(self, url):
        self.info = configparser.ConfigParser()
        req = FetchDistroCore.request(url, False)
        if req is None:
            raise RuntimeError('request failed')
        self.info.read_string(req.text)

    def sha256(self, image):
        line = self.info.get('checksums', image)
        if not line:
            return None
        (algo, checksum) = line.split(':')
        if algo == 'sha256':
            return checksum
        return None

    def release_date(self):
        ts = self.info.get('general', 'timestamp')
        date = datetime.date.fromtimestamp(int(ts))
        return date.isoformat()

    def distro_name(self):
        name = self.info.get('release', 'short').lower()
        version = self.info.get('release', 'version').lower()
        return f'{name}-{version}'

class FetchNetinst(FetchDistroCore):
    """ class for fedora / centos / rhel netinst iso download """

    def __init__(self, mirror, arch = None):
        super().__init__(arch)
        self.mirror = mirror
        # filled by treeinfo()
        self.info = None
        self.netinst = None
        self.outfile = None
        self.image_id = None
        self.checksum = None

    def baseurl(self, _release):
        return None

    def treeinfo(self, release):
        # pylint: disable=assignment-from-none
        base = self.baseurl(release)
        boot = 'images/boot.iso'
        self.info = TreeInfo(f'{base}.treeinfo')
        self.netinst = f'{base}{boot}'
        self.checksum = self.info.sha256(boot)
        reldate = self.info.release_date()
        distname = self.info.distro_name()
        self.outfile = f'{distname}-{self.arch}-netinst-{reldate}.iso'
        self.image_id = f'{distname}-{self.arch}-netinst'

    def fetch_image(self, release, imgdir = 'distro'):
        self.treeinfo(release)
        print(f'# ----- {self.outfile} -----')
        url = self.find_mirror(self.netinst)
        out = f'{imgdir}/{self.outfile}'
        os.makedirs(imgdir, exist_ok = True)
        self.fetch_image_checksum(url, out, self.checksum)
        self.index_update(imgdir, self.image_id, self.outfile, self.checksum,
                          url = url)



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2026 Gerd Hoffmann
#
""" download cloud images """
import os
import sys
import argparse

from virt.qemutest.distro.core import default_imgdir
from virt.qemutest.distro.rhel import FetchRHELCloud
from virt.qemutest.distro.centos import FetchCentosCloud
from virt.qemutest.distro.fedora import FetchFedoraCloud, FetchFedoraUKICloud

def main():
    defdir = default_imgdir()
    parser = argparse.ArgumentParser(description = "download cloud images")
    parser.add_argument('--imgdir', dest = 'imgdir',
                        type = str, default = defdir,
                        help = f'store downloads in DIR (default: {defdir})', metavar = 'DIR')
    parser.add_argument('--fedora', dest = 'fedora',
                        action = 'append', type = str, default = [],
                        help = 'download fedora VER', metavar = 'VER')
    parser.add_argument('--fedora-uki', dest = 'fedora_uki',
                        action = 'append', type = str, default = [],
                        help = 'download fedora VER (UKI variant)', metavar = 'VER')
    parser.add_argument('--centos', dest = 'centos',
                        action = 'append', type = str, default = [],
                        help = 'download centos stream VER', metavar = 'VER')
    parser.add_argument('--rhel', dest = 'rhel',
                        action = 'append', type = str, default = [],
                        help = 'download rhel VER', metavar = 'VER')
    options = parser.parse_args()

    if options.fedora:
        f = FetchFedoraCloud()
        for ver in options.fedora:
            f.fetch_image(ver, imgdir = options.imgdir)

    if options.fedora_uki:
        f = FetchFedoraUKICloud()
        for ver in options.fedora_uki:
            f.fetch_image(ver, imgdir = options.imgdir)

    if options.centos:
        f = FetchCentosCloud()
        for ver in options.centos:
            f.fetch_image(ver, imgdir = options.imgdir)

    if options.rhel:
        m = os.environ.get('REDHAT_MIRROR')
        if m is None:
            m = os.environ.get('BUILDURL') # beaker sets this
        if m is None:
            raise RuntimeError('REDHAT_MIRROR environment variable not set')
        f = FetchRHELCloud(m)
        for ver in options.rhel:
            f.fetch_image(ver, imgdir = options.imgdir)

    return 0

if __name__ == '__main__':
    sys.exit(main())
